@@ -77,7 +77,7 @@ struct buffer *new_outgoing (struct tunnel *t)
 
 inline void recycle_outgoing (struct buffer *buf, struct sockaddr_in peer)
 {
-    /* 
+    /*
      * This should only be used for ZLB's!
      */
     buf->start = buf->rstart + sizeof (struct control_hdr);
@@ -158,6 +158,28 @@ void control_zlb (struct buffer *buf, struct tunnel *t, struct call *c)
     udp_xmit (buf, t);
 }
 
+/*
+ * Get a local address from the local range, if configured.
+ */
+static int get_local_addr(struct tunnel *t, struct call *c)
+{
+#ifdef IP_ALLOCATION
+    if (t->lns->localrange) {
+        c->lns->localaddr = get_addr (t->lns->localrange);
+        if (!c->lns->localaddr)
+        {
+            set_error (c, ERROR_NORES, "No available local IP addresses");
+            call_close (c);
+            l2tp_log (LOG_DEBUG, "%s: Out of local IP addresses on tunnel %d!\n",
+                     __FUNCTION__, t->tid);
+            return -EINVAL;
+        }
+        reserve_addr (c->lns->localaddr);
+    }
+#endif
+    return 0;
+}
+
 int control_finish (struct tunnel *t, struct call *c)
 {
     /*
@@ -179,7 +201,7 @@ int control_finish (struct tunnel *t, struct call *c)
     char dummy_buf[128] = "/var/l2tp/"; /* jz: needed to read /etc/ppp/var.options - just kick it if you don't like */
     char passwdfd_buf[32] = ""; /* buffer for the fd, not the password */
     int i;
-    int pppd_passwdfd[2];            
+    int pppd_passwdfd[2];
     int tmptid,tmpcid;
 
     if (c->msgtype < 0)
@@ -392,7 +414,7 @@ int control_finish (struct tunnel *t, struct call *c)
         /* FIXME: Do we need to be sure they specified a version number?
          *   Theoretically, yes, but we don't have anything in the code
          *   to actually *do* anything with it, so...why check at this point?
-         * We shouldn't be requiring a bearer capabilities avp to be present in 
+         * We shouldn't be requiring a bearer capabilities AVP to be present in
          * SCCRQ and SCCRP as they aren't required
          if (t->bc < 0 ) {
          if (DEBUG) l2tp_log(LOG_DEBUG,
@@ -420,7 +442,7 @@ int control_finish (struct tunnel *t, struct call *c)
                 (y != t))
             {
                 /* This can happen if we get a duplicate
-                   StartCCN or if they don't get our ack packet */
+                   StartCCN or if they don't get our ACK packet */
                 /*
                  * But it is legitimate for two different remote systems
                  * to use the same tid
@@ -516,7 +538,7 @@ int control_finish (struct tunnel *t, struct call *c)
         /* FIXME: Do we need to be sure they specified a version number?
          *   Theoretically, yes, but we don't have anything in the code
          *   to actually *do* anything with it, so...why check at this point?
-         * We shouldn't be requiring a bearer capabilities avp to be present in 
+         * We shouldn't be requiring a bearer capabilities AVP to be present in
          * SCCRQ and SCCRP as they aren't required
          if (t->bc < 0 ) {
          if (DEBUG) log(LOG_DEBUG,
@@ -713,7 +735,7 @@ int control_finish (struct tunnel *t, struct call *c)
                 l2tp_log (LOG_DEBUG,
                      "%s: Peer tried to initiate call without call ID\n",
                      __FUNCTION__);
-            /* Here it doesn't make sense to use the needclose flag because 
+            /* Here it doesn't make sense to use the needclose flag because
                the call p did not receive any packets */
             call_close (p);
             return -EINVAL;
@@ -724,7 +746,7 @@ int control_finish (struct tunnel *t, struct call *c)
             if (z->cid == p->cid)
             {
                 /* This can happen if we get a duplicate
-                   ICRQ or if they don't get our ack packet */
+                   ICRQ or if they don't get our ACK packet */
                 l2tp_log (LOG_DEBUG,
                      "%s: Peer requested call %d twice, ignoring second one.\n",
                      __FUNCTION__, p->cid);
@@ -877,7 +899,7 @@ int control_finish (struct tunnel *t, struct call *c)
             if (c->lac->debug)
                 po = add_opt (po, "debug");
             if (c->lac->password[0])
-            {                    
+            {
                 if (pipe (pppd_passwdfd) == -1)
                 {
                   l2tp_log (LOG_DEBUG,
@@ -949,6 +971,8 @@ int control_finish (struct tunnel *t, struct call *c)
             */
         }
         c->state = ICCN;
+        if (get_local_addr(t, c))
+            return -EINVAL;
         strncpy (ip1, IPADDY (c->lns->localaddr), sizeof (ip1));
         strncpy (ip2, IPADDY (c->addr), sizeof (ip2));
         po = NULL;
@@ -994,11 +1018,12 @@ int control_finish (struct tunnel *t, struct call *c)
             po = add_opt (po, "ipparam");
             po = add_opt (po, IPADDY (t->peer.sin_addr));
         }
+
         start_pppd (c, po);
         opt_destroy (po);
         l2tp_log (LOG_NOTICE,
-             "Call established with %s, Local: %d, Remote: %d, Serial: %d\n",
-             IPADDY (t->peer.sin_addr), c->ourcid, c->cid,
+             "Call established with %s, PID: %d, Local: %d, Remote: %d, Serial: %d\n",
+             IPADDY (t->peer.sin_addr), c->pppd, c->ourcid, c->cid,
              c->serno);
         break;
     case OCRP:                 /* jz: nothing to do for OCRP, waiting for OCCN */
@@ -1062,8 +1087,8 @@ int control_finish (struct tunnel *t, struct call *c)
         /*  jz: just show some information */
         l2tp_log (LOG_INFO,
 		  "parameters: Local: %d , Remote: %d , Serial: %d , Pid: %d , Tunnelid: %d , Phoneid: %s\n",
-		  c->ourcid, c->cid, c->serno, c->pppd, t->ourtid, c->dial_no); 
-	
+		  c->ourcid, c->cid, c->serno, c->pppd, t->ourtid, c->dial_no);
+
         opt_destroy (po);
         if (c->lac)
             c->lac->rtries = 0;
@@ -1139,7 +1164,7 @@ int control_finish (struct tunnel *t, struct call *c)
     return 0;
 }
 
-inline int check_control (const struct buffer *buf, struct tunnel *t,
+static inline int check_control (const struct buffer *buf, struct tunnel *t,
                           struct call *c)
 {
     /*
@@ -1181,7 +1206,7 @@ inline int check_control (const struct buffer *buf, struct tunnel *t,
             l2tp_log (LOG_DEBUG,
                  "%s: Received out of order control packet on tunnel %d (got %d, expected %d)\n",
                  __FUNCTION__, t->tid, h->Ns, t->control_rec_seq_num);
-        if (((h->Ns < t->control_rec_seq_num) && 
+        if (((h->Ns < t->control_rec_seq_num) &&
             ((t->control_rec_seq_num - h->Ns) < 32768)) ||
             ((h->Ns > t->control_rec_seq_num) &&
             ((t->control_rec_seq_num - h->Ns) > 32768)))
@@ -1196,10 +1221,14 @@ inline int check_control (const struct buffer *buf, struct tunnel *t,
                 l2tp_log (LOG_DEBUG, "%s: Sending an updated ZLB in reponse\n",
                      __FUNCTION__);
 #endif
-            zlb = new_outgoing (t);
-            control_zlb (zlb, t, c);
-            udp_xmit (zlb, t);
-            toss (zlb);
+            if (buf->len != sizeof (struct control_hdr))
+            {
+                /* don't send a ZLB in response to a ZLB. it leads to a loop */
+                zlb = new_outgoing (t);
+                control_zlb (zlb, t, c);
+                /*udp_xmit (zlb, t);*/
+                toss (zlb);
+            }
         }
         else if (!t->control_rec_seq_num && (t->tid == -1))
         {
@@ -1275,7 +1304,7 @@ inline int check_control (const struct buffer *buf, struct tunnel *t,
     return 0;
 }
 
-inline int check_payload (struct buffer *buf, struct tunnel *t,
+static inline int check_payload (struct buffer *buf, struct tunnel *t,
                           struct call *c)
 {
     /*
@@ -1289,18 +1318,18 @@ inline int check_payload (struct buffer *buf, struct tunnel *t,
     {
         if (DEBUG)
         {
-            l2tp_log (LOG_DEBUG, "%s: Aempted to send payload on tunnel\n",
+            l2tp_log (LOG_DEBUG, "%s: Attempted to send payload on tunnel\n",
                  __FUNCTION__);
         }
         return -EINVAL;
     }
     if (buf->len < MIN_PAYLOAD_HDR_LEN)
     {
-        /* has to be at least MIN_PAYLOAD_HDR_LEN 
+        /* has to be at least MIN_PAYLOAD_HDR_LEN
            no matter what.  we'll look more later */
         if (DEBUG)
         {
-            l2tp_log (LOG_DEBUG, "%s:Recieved to small of packet\n", __FUNCTION__);
+            l2tp_log (LOG_DEBUG, "%s:Received to small of packet\n", __FUNCTION__);
         }
         return -EINVAL;
     }
@@ -1381,7 +1410,7 @@ inline int check_payload (struct buffer *buf, struct tunnel *t,
 #endif
     return 0;
 }
-inline int expand_payload (struct buffer *buf, struct tunnel *t,
+static inline int expand_payload (struct buffer *buf, struct tunnel *t,
                            struct call *c)
 {
     /*
@@ -1477,9 +1506,9 @@ inline int expand_payload (struct buffer *buf, struct tunnel *t,
     if (new_hdr->Ns != c->data_seq_num)
     {
         /* RFC1982-esque comparison of serial numbers */
-        if (((new_hdr->Ns < c->data_rec_seq_num) && 
+        if (((new_hdr->Ns < c->data_rec_seq_num) &&
             ((c->data_rec_seq_num - new_hdr->Ns) < 32768)) ||
-            ((new_hdr->Ns > c->data_rec_seq_num) && 
+            ((new_hdr->Ns > c->data_rec_seq_num) &&
             ((c->data_rec_seq_num - new_hdr->Ns) > 32768)))
         {
 #ifdef DEBUG_FLOW
@@ -1548,7 +1577,7 @@ void send_zlb (void *data)
              __FUNCTION__);
         return;
     }
-    /* Update the counter so we know what Lr was when we last transmited a ZLB */
+    /* Update the counter so we know what Lr was when we last transmitted a ZLB */
     c->prx = c->data_rec_seq_num;
     buf = new_payload (t->peer);
     add_payload_hdr (t, c, buf);
@@ -1561,7 +1590,7 @@ void send_zlb (void *data)
     toss (buf);
 }
 
-inline int write_packet (struct buffer *buf, struct tunnel *t, struct call *c,
+static inline int write_packet (struct buffer *buf, struct tunnel *t, struct call *c,
                          int convert)
 {
     /*
@@ -1576,12 +1605,12 @@ inline int write_packet (struct buffer *buf, struct tunnel *t, struct call *c,
 
     if (c->fd < 0)
     {
-        if (DEBUG || 1)
+        if (DEBUG)
             l2tp_log (LOG_DEBUG, "%s: tty is not open yet.\n", __FUNCTION__);
         return -EIO;
     }
     /*
-     * Skip over header 
+     * Skip over header
      */
     _u16 offset = ((struct payload_hdr*)(buf->start))->o_size;  // For FIXME:
     buf->start += sizeof(struct payload_hdr) + offset;
@@ -1636,7 +1665,7 @@ inline int write_packet (struct buffer *buf, struct tunnel *t, struct call *c,
 
     /*
      * sync->async conversion if we're doing sync frames
-     * since the pppd driver will expect async frames
+     * since the PPPD driver will expect async frames
      * Write leading flag character
      */
 
@@ -1697,7 +1726,7 @@ inline int write_packet (struct buffer *buf, struct tunnel *t, struct call *c,
     return 0;
 }
 
-void handle_special (struct buffer *buf, struct call *c, _u16 call)
+int handle_special (struct buffer *buf, struct call *c, _u16 call)
 {
     /*
        * This procedure is called when we have received a packet
@@ -1709,7 +1738,7 @@ void handle_special (struct buffer *buf, struct call *c, _u16 call)
     struct tunnel *t = c->container;
     /* Don't do anything unless it's a control packet */
     if (!CTBIT (*((_u16 *) buf->start)))
-        return;
+        return 0;
     /* Temporarily, we make the tunnel have cid of call instead of 0,
        but we need to stop any scheduled events (like Hello's in
        particular) which might use this value */
@@ -1722,14 +1751,15 @@ void handle_special (struct buffer *buf, struct call *c, _u16 call)
             if (gconfig.debug_tunnel)
                 l2tp_log (LOG_DEBUG, "%s: ZLB for closed call\n", __FUNCTION__);
             c->cid = 0;
-            return;
+            return 0;
         }
         /* Make a packet with the specified call number */
         /* FIXME: If I'm not a CDN, I need to send a CDN */
         control_zlb (buf, t, c);
         c->cid = 0;
-        udp_xmit (buf, t);
+        /*udp_xmit (buf, t);*/
         toss (buf);
+        return 1;
     }
     else
     {
@@ -1737,6 +1767,7 @@ void handle_special (struct buffer *buf, struct call *c, _u16 call)
         if (gconfig.debug_tunnel)
             l2tp_log (LOG_DEBUG, "%s: invalid control packet\n", __FUNCTION__);
     }
+    return 0;
 }
 
 inline int handle_packet (struct buffer *buf, struct tunnel *t,
@@ -1818,10 +1849,10 @@ inline int handle_packet (struct buffer *buf, struct tunnel *t,
                     res = write_packet (buf, t, c, SYNC_FRAMING);
                     if (res)
                         return res;
-                    /* 
+                    /*
                        * Assuming we wrote to the ppp driver okay, we should
                        * do something about ZLB's unless *we* requested no
-                       * window size or if they we have turned off our fbit. 
+                       * window size or if they we have turned off our fbit.
                      */
 
 /*					if (c->ourfbit && (c->ourrws > 0)) {
