@@ -20,6 +20,8 @@ typedef unsigned long long _u64;
 #define _L2TP_H
 
 #define MAXSTRLEN 120           /* Maximum length of common strings */
+                                /* FIXME: MAX_RECV_SIZE, what is it? */
+#define MAX_RECV_SIZE 4096      /* Biggest packet we'll accept */
 
 #include <netinet/in.h>
 #include <termios.h>
@@ -40,8 +42,23 @@ typedef unsigned long long _u64;
 #define CONTROL_PIPE "/var/run/xl2tpd/l2tp-control"
 #define CONTROL_PIPE_MESSAGE_SIZE 1024
 
+/* Control pip request types */
+#define CONTROL_PIPE_REQ_LAC_REMOVE 'r'
+#define CONTROL_PIPE_REQ_LAC_ADD_MODIFY 'a'
+#define CONTROL_PIPE_REQ_LAC_STATUS 's'
+#define CONTROL_PIPE_REQ_LAC_DISCONNECT 'd'
+#define CONTROL_PIPE_REQ_LAC_HANGUP 'h'
+#define CONTROL_PIPE_REQ_LAC_OUTGOING_CALL 'o'
+#define CONTROL_PIPE_REQ_LAC_CONNECT 'c'
+#define CONTROL_PIPE_REQ_TUNNEL 't'
+
+#define CONTROL_PIPE_REQ_LNS_ADD_MODIFY 'z' /* Create or modify an existing LNS */
+#define CONTROL_PIPE_REQ_LNS_STATUS 'y'     /* Get status of LNS */
+#define CONTROL_PIPE_REQ_AVAILABLE 'x'     /* Get status of LNS */
+#define CONTROL_PIPE_REQ_LNS_REMOVE 'w'     /* Get status of LNS */
+
 #define BINARY "xl2tpd"
-#define SERVER_VERSION "xl2tpd-1.3.6"
+#define SERVER_VERSION "xl2tpd-1.3.11"
 #define VENDOR_NAME "xelerance.com"
 #ifndef PPPD
 #define PPPD		"/usr/sbin/pppd"
@@ -65,7 +82,7 @@ struct control_hdr
 #define CLBIT(ver) (ver & 0x4000)       /* Length bit present.  Must be 1
                                            for control messages */
 
-#define CZBITS(ver) (ver &0x37F8)       /* Reserved bits:  We must drop 
+#define CZBITS(ver) (ver &0x37F8)       /* Reserved bits:  We must drop
                                            anything with these there */
 
 #define CFBIT(ver) (ver & 0x0800)       /* Presence of Ns and Nr fields
@@ -94,12 +111,12 @@ struct payload_hdr
                                    so we can "decompress" the
                                    payloads and simplify coding */
 #if 1
-#define DEFAULT_MAX_RETRIES 5    /* Recommended value from spec */
+#define DEFAULT_MAX_RETRIES 5   /* Recommended value from spec */
 #else
 #define DEFAULT_MAX_RETRIES 95   /* give us more time to debug */
 #endif
 
-#define DEFAULT_RWS_SIZE   4    /* Default max outstanding 
+#define DEFAULT_RWS_SIZE   4    /* Default max outstanding
                                    control packets in queue */
 #define DEFAULT_TX_BPS		10000000        /* For outgoing calls, report this speed */
 #define DEFAULT_RX_BPS		10000000
@@ -109,8 +126,6 @@ struct payload_hdr
 #define MIN_PAYLOAD_HDR_LEN 6
 
 #define UDP_LISTEN_PORT  1701
-                                /* FIXME: MAX_RECV_SIZE, what is it? */
-#define MAX_RECV_SIZE 4096      /* Biggest packet we'll accept */
 
 #define OUR_L2TP_VERSION 0x100  /* We support version 1, revision 0 */
 
@@ -155,11 +170,6 @@ struct tunnel
     _u16 control_seq_num;       /* Sequence for next packet */
     _u16 control_rec_seq_num;   /* Next expected to receive */
     int cLr;                    /* Last packet received by peer */
-    char hostname[MAXSTRLEN];   /* Remote hostname */
-    char vendor[MAXSTRLEN];     /* Vendor of remote product */
-    struct challenge chal_us;   /* Their Challenge to us */
-    struct challenge chal_them; /* Our challenge to them */
-    char secret[MAXSTRLEN];     /* Secret to use */
 #ifdef SANITY
     int sanity;                 /* check for sanity? */
 #endif
@@ -173,6 +183,11 @@ struct tunnel
     struct lns *lns;            /* LNS that owns us */
     struct lac *lac;            /* LAC that owns us */
     struct in_pktinfo my_addr;  /* Address of my endpoint */
+    char hostname[MAXSTRLEN];   /* Remote hostname */
+    char vendor[MAXSTRLEN];     /* Vendor of remote product */
+    struct challenge chal_us;   /* Their Challenge to us */
+    struct challenge chal_them; /* Our challenge to them */
+    char secret[MAXSTRLEN];     /* Secret to use */
 };
 
 struct tunnel_list
@@ -216,7 +231,7 @@ extern void destroy_tunnel (struct tunnel *);
 extern struct buffer *new_payload (struct sockaddr_in);
 extern void recycle_payload (struct buffer *, struct sockaddr_in);
 extern void add_payload_hdr (struct tunnel *, struct call *, struct buffer *);
-extern int read_packet (struct buffer *, int, int);
+extern int read_packet (struct call *);
 extern void udp_xmit (struct buffer *buf, struct tunnel *t);
 extern void control_xmit (void *);
 extern int ppd;
@@ -236,7 +251,7 @@ extern int get_entropy (unsigned char *, int);
 #endif
 
 
-/* 
+/*
  * This is just some stuff to take
  * care of kernel definitions
  */
